/*-
 * Copyright (c) 2003-2004 Tim Kientzle
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "archive_platform.h"
__FBSDID("$FreeBSD: src/lib/libarchive/archive_write_set_format_cpio.c,v 1.4.2.1 2004/12/22 00:01:55 kientzle Exp $");

#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <archive.h>
#include <archive_entry.h>
#include <archive_private.h>
 
#include <tpf/tpf_tar.h>

#define etoa( a, b)                      \
  {                                      \
    int cnt = b;                         \
    u_char *l = (u_char *)a;             \
    for (; cnt--; ++l)                   \
      *l = e2ASCIItab[*l];               \
  }

 
static int	archive_write_cpio_data(struct archive *, const void *buff,
		    size_t s);
static int	archive_write_cpio_finish(struct archive *);
static int	archive_write_cpio_finish_entry(struct archive *);
static int	archive_write_cpio_header(struct archive *,
		    struct archive_entry *);
static int	format_octal(int64_t, void *, int);
static int64_t	format_octal_recursive(int64_t, char *, int);

struct cpio {
	uint64_t	  entry_bytes_remaining;
};

struct cpio_header {
	char	c_magic[6];
	char	c_dev[6];
	char	c_ino[6];
	char	c_mode[6];
	char	c_uid[6];
	char	c_gid[6];
	char	c_nlink[6];
	char	c_rdev[6];
	char	c_mtime[11];
	char	c_namesize[6];
	char	c_filesize[11];
};


/* EBCDIC to ASCII -- POSIX and System V compatible. */
static u_char e2ASCIItab[] = {
  0000, 0001, 0002, 0003, 0234, 0011, 0206, 0177,    /* 0000 */
  0227, 0215, 0216, 0013, 0014, 0015, 0016, 0017,    /* 0010 */
  0020, 0021, 0022, 0023, 0235, 0205, 0010, 0207,    /* 0020 */
  0030, 0031, 0222, 0217, 0034, 0035, 0036, 0037,    /* 0030 */
  0200, 0201, 0202, 0203, 0204, 0012, 0027, 0033,    /* 0040 */
  0210, 0211, 0212, 0213, 0214, 0005, 0006, 0007,    /* 0050 */
  0220, 0221, 0026, 0223, 0224, 0225, 0226, 0004,    /* 0060 */
  0230, 0231, 0232, 0233, 0024, 0025, 0236, 0032,    /* 0070 */
  0040, 0240, 0241, 0242, 0243, 0244, 0245, 0246,    /* 0100 */
  0247, 0250, 0325, 0056, 0074, 0050, 0053, 0174,    /* 0110 */
  0046, 0251, 0252, 0253, 0254, 0255, 0256, 0257,    /* 0120 */
  0260, 0261, 0041, 0044, 0052, 0051, 0073, 0176,    /* 0130 */
  0055, 0057, 0262, 0263, 0264, 0265, 0266, 0267,    /* 0140 */
  0270, 0271, 0313, 0054, 0045, 0137, 0076, 0077,    /* 0150 */
  0272, 0273, 0274, 0275, 0276, 0277, 0300, 0301,    /* 0160 */
  0302, 0140, 0072, 0043, 0100, 0047, 0075, 0042,    /* 0170 */
  0303, 0141, 0142, 0143, 0144, 0145, 0146, 0147,    /* 0200 */
  0150, 0151, 0304, 0305, 0306, 0307, 0310, 0311,    /* 0210 */
  0312, 0152, 0153, 0154, 0155, 0156, 0157, 0160,    /* 0220 */
  0161, 0162, 0136, 0314, 0315, 0316, 0317, 0320,    /* 0230 */
  0321, 0345, 0163, 0164, 0165, 0166, 0167, 0170,    /* 0240 */
  0171, 0172, 0322, 0323, 0324, 0133, 0326, 0327,    /* 0250 */
  0330, 0331, 0332, 0333, 0334, 0335, 0336, 0337,    /* 0260 */
  0340, 0341, 0342, 0343, 0344, 0135, 0346, 0347,    /* 0270 */
  0173, 0101, 0102, 0103, 0104, 0105, 0106, 0107,    /* 0300 */
  0110, 0111, 0350, 0351, 0352, 0353, 0354, 0355,    /* 0310 */
  0175, 0112, 0113, 0114, 0115, 0116, 0117, 0120,    /* 0320 */
  0121, 0122, 0356, 0357, 0360, 0361, 0362, 0363,    /* 0330 */
  0134, 0237, 0123, 0124, 0125, 0126, 0127, 0130,    /* 0340 */
  0131, 0132, 0364, 0365, 0366, 0367, 0370, 0371,    /* 0350 */
  0060, 0061, 0062, 0063, 0064, 0065, 0066, 0067,    /* 0360 */
  0070, 0071, 0372, 0373, 0374, 0375, 0376, 0377,    /* 0370 */
};

 
/*
 * Set output format to 'cpio' format.
 */
int
archive_write_set_format_ASCIIcpio(struct archive *a)
{
	struct cpio *cpio;

	/* If someone else was already registered, unregister them. */
	if (a->format_finish != NULL)
		(a->format_finish)(a);

	cpio = malloc(sizeof(*cpio));
	if (cpio == NULL) {
		archive_set_error(a, ENOMEM, "Can't allocate cpio data");
		return (ARCHIVE_FATAL);
	}
	memset(cpio, 0, sizeof(*cpio));
	a->format_data = cpio;

	a->pad_uncompressed = 1;
	a->format_write_header = archive_write_cpio_header;
	a->format_write_data = archive_write_cpio_data;
	a->format_finish_entry = archive_write_cpio_finish_entry;
	a->format_finish = archive_write_cpio_finish;
	a->archive_format = ARCHIVE_FORMAT_CPIO_ASCII;
	a->archive_format_name = "POSIX cpio";
	return (ARCHIVE_OK);
}

static int
archive_write_cpio_header(struct archive *a, struct archive_entry *entry)
{
	struct cpio *cpio;
	const char *p, *path;
	char *path1;
	int pathlength, ret;
	const struct stat	*st;
	struct cpio_header	 h;

	cpio = a->format_data;
	ret = 0;

	path = archive_entry_pathname(entry);
	pathlength = strlen(path) + 1; /* Include trailing null. */
	st = archive_entry_stat(entry);

	memset(&h, 0, sizeof(h));
	format_octal(070707,
	             &h.c_magic, sizeof(h.c_magic));
	format_octal(st->st_dev, &h.c_dev, sizeof(h.c_dev));
	/*
	 * TODO: Generate artificial inode numbers rather than just
	 * re-using the ones off the disk.  That way, the 18-bit c_ino
	 * field only limits the number of files in the archive.
	 */
	if (st->st_ino > 0777777) {
		archive_set_error(a, ERANGE, "large inode number truncated");
		ret = ARCHIVE_WARN;
	}

	format_octal(st->st_ino & 0777777, &h.c_ino, sizeof(h.c_ino));
        {
           int work;
           mode32to16(st->st_mode, work)
           work = convertTPFtoBSDfiletype(work);
           format_octal(work, &h.c_mode, sizeof(h.c_mode));
        }
	format_octal(st->st_uid, &h.c_uid, sizeof(h.c_uid));
	format_octal(st->st_gid, &h.c_gid, sizeof(h.c_gid));
	format_octal(st->st_nlink, &h.c_nlink, sizeof(h.c_nlink));
	if(S_ISBLK(st->st_mode) || S_ISCHR(st->st_mode))
	    format_octal(st->st_rdev, &h.c_rdev, sizeof(h.c_rdev));
	else
	    format_octal(0, &h.c_rdev, sizeof(h.c_rdev));
	format_octal(st->st_mtime, &h.c_mtime, sizeof(h.c_mtime));
	format_octal(pathlength, &h.c_namesize, sizeof(h.c_namesize));

	/* Symlinks get the link written as the body of the entry. */
	p = archive_entry_symlink(entry);
	if (p != NULL  &&  *p != '\0')
		format_octal(strlen(p), &h.c_filesize, sizeof(h.c_filesize));
	else
		format_octal(st->st_size, &h.c_filesize, sizeof(h.c_filesize));

	ret = (a->compression_write)(a, &h, sizeof(h));
	if (ret != ARCHIVE_OK)
		return (ARCHIVE_FATAL);

	path1 = malloc(pathlength);
	memcpy(path1, path, pathlength);
	etoa(path1, pathlength);
 
	ret = (a->compression_write)(a, path1, pathlength);
	free(path1);
 
	if (ret != ARCHIVE_OK)
		return (ARCHIVE_FATAL);

	cpio->entry_bytes_remaining = st->st_size;

	/* Write the symlink now. */
	if (p != NULL  &&  *p != '\0')
		ret = (a->compression_write)(a, p, strlen(p));

	return (ret);
}

static int
archive_write_cpio_data(struct archive *a, const void *buff, size_t s)
{
	struct cpio *cpio;
	int ret;

	cpio = a->format_data;
	if (s > cpio->entry_bytes_remaining)
		s = cpio->entry_bytes_remaining;

	ret = (a->compression_write)(a, buff, s);
	cpio->entry_bytes_remaining -= s;
	return (ret);
}

/*
 * Format a number into the specified field.
 */
static int
format_octal(int64_t v, void *p, int digits)
{
	int64_t	max;
	int	ret;

	max = (((int64_t)1) << (digits * 3)) - 1;
	if (v >= 0  &&  v <= max) {
	    format_octal_recursive(v, p, digits);
	    ret = 0;
	} else {
	    format_octal_recursive(max, p, digits);
	    ret = -1;
	}
	return (ret);
}

static int64_t
format_octal_recursive(int64_t v, char *p, int s)
{
	if (s == 0)
		return (v);
	v = format_octal_recursive(v, p+1, s-1);
	*p = 0x30 + (v & 7);
	return (v >>= 3);
}

static int
archive_write_cpio_finish(struct archive *a)
{
	struct cpio *cpio;
	struct stat st;
	int er;
	struct archive_entry *trailer;

	cpio = a->format_data;
	trailer = archive_entry_new();
	memset(&st, 0, sizeof(st));
	st.st_nlink = 1;
	archive_entry_copy_stat(trailer, &st);
	archive_entry_set_pathname(trailer,
	  "\x54\x52\x41\x49\x4c\x45\x52\x21\x21\x21"); /* "TRAILER!!!" */
	er = archive_write_cpio_header(a, trailer);
	archive_entry_free(trailer);

	free(cpio);
	a->format_data = NULL;
	return (er);
}

static int
archive_write_cpio_finish_entry(struct archive *a)
{
	struct cpio *cpio;
	int to_write, ret;

	cpio = a->format_data;
	ret = ARCHIVE_OK;
	while (cpio->entry_bytes_remaining > 0) {
		to_write = cpio->entry_bytes_remaining < a->null_length ?
		    cpio->entry_bytes_remaining : a->null_length;
		ret = (a->compression_write)(a, a->nulls, to_write);
		if (ret != ARCHIVE_OK)
			return (ret);
		cpio->entry_bytes_remaining -= to_write;
	}
	return (ret);
}
